/*
 * Copyright (c) 1993-1995 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Contributed by Bob Olson (olson@mcs.anl.gov) September 1995.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: grabber-win32.cc,v 1.0 96/10/25 13:38:38 mccanne Exp $ (LBL)";
#endif

#include <stdio.h>
#include <errno.h>
#include <string.h>
#ifndef WIN32
#include <sys/file.h>
#endif
#include <sys/stat.h>
    
#include "inet.h"
#include "Tcl.h"
#include "rtp.h"
#include "grabber.h"
#include "crdef.h"
#include "device-input.h"
#include "transmitter.h"
#include "module.h"
#include "rgb-converter.h"

#define NOCOMPMAN
#define NODRAWDIB
#define NOAVIFMT
#define NOAVIFILE
#define NOMCIWND
#define NOMSACM
#include <vfw.h>

class WinGrabber : public Grabber {
public:
	WinGrabber();
	virtual ~WinGrabber();
	virtual int command(int argc, const char* const * argv);
	virtual void setsize()=0;
	virtual void start();
	virtual void stop();

	void getframe(LPVIDEOHDR lphd);
	BOOL sourcedlg;

protected:
	void startup();
	void shutdown();
	int grab();

	void askimagesize();
	void setdibsize(int w,int h);

	int decimate_;
	RGB_Converter* converter_;

	char *dib;
	BITMAPINFOHEADER bi;
	HWND capwin;
	int imw,imh;
};

class Win422Grabber : public WinGrabber {
public:
	virtual void setsize();
};

class Win411Grabber : public WinGrabber {
public:
	virtual void setsize();
};

class WinCIFGrabber : public WinGrabber {
public:
	virtual void setsize();
};

class CapDevice : public InputDevice {
public:
	WinGrabber *last_grabber;
	char *capName;
	CapDevice(const char* s);
	~CapDevice();
	virtual int command(int argc, const char * const * argv);
};

static CapDevice cap_device("msvideo");

CapDevice::CapDevice(const char* s)
:InputDevice(s),capName(0),last_grabber(0)
{
	char name[70];
	char ver[20];
	if (capGetDriverDescription(0,name,40,ver,20)) {
		if (ver[0]) {
			strcat(name," : ");
			strcat(name,ver);
		}
		capName=strdup(name);
		attributes_ = "format { 422 411 } size { small large cif }";
	}
	else attributes_ = "disabled";
}

CapDevice::~CapDevice()
{
	if (last_grabber) {
		last_grabber->sourcedlg=0;
		delete last_grabber;
	}
}

int CapDevice::command(int argc, const char*const* argv)
{
	if (argc == 2) {
		if (strcmp(argv[1], "attributes") == 0) {
			Tcl_SetVar(Tcl::instance().interp(),
				"msvideoName",capName?capName:"(no device)", TCL_GLOBAL_ONLY);			
			Tcl::instance().result(attributes_);
			return (TCL_OK);
		}
	} else
	if (argc == 3)
	{
		if (strcmp(argv[1], "open") == 0)
		{
			const char* fmt = argv[2];
			if (last_grabber) delete last_grabber;
			last_grabber=0;
			if (strcmp(fmt, "422") == 0)
				last_grabber = new Win422Grabber();
			else if (strcmp(fmt, "411") == 0)
				last_grabber = new Win411Grabber();
			else if (strcmp(fmt, "cif") == 0)
				last_grabber = new WinCIFGrabber();
			if (last_grabber != 0) {
				Tcl::instance().result(last_grabber->name());
			}
			return (TCL_OK);
		}
	}
	return (InputDevice::command(argc, argv));
}

WinGrabber::WinGrabber() :
	decimate_(2), dib(0), capwin(0), sourcedlg(0)
{
}

WinGrabber::~WinGrabber()
{
	shutdown();
	cap_device.last_grabber=0;
}

int WinGrabber::command(int argc, const char * const * argv)
{
	Tcl& tcl = Tcl::instance();
	if (argc == 3)
	{
		if (strcmp(argv[1], "decimate") == 0) {
			int d = atoi(argv[2]);
			if (d <= 0) {
				Tcl_AppendResult(tcl.interp(),
					"%s: divide by zero",
					argv[0], 0);
				return TCL_ERROR;
			}
			if (d != decimate_) {
				decimate_ = d;
				startup();
			}
			return (TCL_OK);
		}
	} 
	if (strcmp(argv[1], "setsource") == 0) {
		if (sourcedlg) {
			capDlgVideoSource(capwin);
			startup();
		}
		return (TCL_OK);
	}
	return (Grabber::command(argc, argv));
}

LRESULT FAR PASCAL FrameCallback (HWND capwin, LPVIDEOHDR lphd)
{
	((WinGrabber*)(capGetUserData(capwin)))->getframe(lphd);
	 return (LRESULT) TRUE;
}

void WinGrabber::startup()
{
	shutdown();
	capwin=capCreateCaptureWindow("WinCap",WS_OVERLAPPED,0,0,32,32,0,0);
	if (!capwin) goto failed;
	if (!capDriverConnect(capwin,0)) goto failed;
	CAPDRIVERCAPS caps;
	if (!capDriverGetCaps (capwin,&caps,sizeof(caps))) goto failed;
	if (!caps.fCaptureInitialized) goto failed;
	sourcedlg=caps.fHasDlgVideoSource;

	setsize();

	bi.biSize            = sizeof(BITMAPINFOHEADER);
	bi.biCompression     = BI_RGB;
	bi.biWidth           = outw_;
	bi.biHeight          = outh_;
	bi.biPlanes          = 1;
	bi.biSizeImage       = 0;
	bi.biXPelsPerMeter   = 0;
	bi.biYPelsPerMeter   = 0;
	bi.biClrUsed         = 0;
	bi.biClrImportant    = 0;

	bi.biBitCount        = 32;
	if (capSetVideoFormat(capwin,&bi,sizeof(bi))) goto setted;
	bi.biBitCount        = 24;
	if (capSetVideoFormat(capwin,&bi,sizeof(bi))) goto setted;
	bi.biBitCount        = 16;
	if (capSetVideoFormat(capwin,&bi,sizeof(bi))) goto setted;

	/* not rgb format, decompressor needed*/
	/* not implemented */
	goto failed; 

setted:

	if (!capSetCallbackOnFrame (capwin,FrameCallback)) goto failed;
	if (!capSetUserData(capwin,this)) goto failed;
	
	Tcl::instance().evalc(
		sourcedlg?"$msvideoSource configure -state normal"
					:"$msvideoSource configure -state disabled");

	return;
failed:
	shutdown();
	status_=-1; 
	printf("grabber startup failed");
}

void WinGrabber::shutdown()
{
	if (sourcedlg)
		Tcl::instance().evalc("$msvideoSource configure -state disabled");
	sourcedlg=0;
	if (capwin) DestroyWindow(capwin);
	capwin=0;
	if (dib) delete dib;
	dib=0;
}

void WinGrabber::start()
{
	startup();
	Grabber::start();
}

void WinGrabber::stop()
{
	shutdown();
	Grabber::stop();
}

void WinGrabber::getframe(LPVIDEOHDR lphd)
{
	if (!dib) return;
	if (bi.biCompression==BI_RGB) {
		if (bi.biBitCount==32) {
			char *v=(char*)(lphd->lpData);
			for (int i=outh_-1; i>=0; i--) {
				char *p=&dib[i*outw_*4];
				for (int j=outw_; j>0; j--) {
					p[0]=v[2];
					p[1]=v[1];
					p[2]=v[0];
					p[3]=0;
					p+=4;
					v+=4;
				}
			}
		}
		if (bi.biBitCount==24) {
			char *v=(char*)(lphd->lpData);
			for (int i=outh_-1; i>=0; i--) {
				char *p=&dib[i*outw_*4];
				for (int j=outw_; j>0; j--) {
					p[0]=v[2];
					p[1]=v[1];
					p[2]=v[0];
					p[3]=0;
					p+=4;
					v+=3;
				}
			}
		}
		if (bi.biBitCount==16) {
			WORD *v=(WORD*)(lphd->lpData);
			for (int i=outh_-1; i>=0; i--) {
				char *p=&dib[i*outw_*4];
				for (int j=outw_; j>0; j--) {
					p[0]=(*v>>7)&0xfc;
					p[1]=(*v>>2)&0xfc;
					p[2]=(*v<<3)&0xfc;
					p[3]=0;
					p+=4;
					v+=1;
				}
			}
		}
	}
}

static int cangrab=1;
void LetGrab(){cangrab=1;}

int WinGrabber::grab()
{
	if (dib && cangrab) {
		cangrab=0;
		capGrabFrame(capwin);
		converter_->convert((u_int8_t*)dib,outw_,outh_,frame_);
		suppress(frame_);
		saveblks(frame_);
		YuvFrame f(media_ts(), frame_, crvec_, outw_, outh_);
		target_->consume(&f);
	}
	return 1;
}

void WinGrabber::askimagesize()
{
	CAPSTATUS st;
	if (capwin && capGetStatus(capwin,&st,sizeof(st))) {
		imw=st.uiImageWidth;
		imh=st.uiImageHeight;
	} else {
		imw=640;
		imh=480;
	}
}

void WinGrabber::setdibsize(int w,int h)
{
	if (dib) delete dib;
	dib=new char[w*h*4];
}

void Win422Grabber::setsize()
{
	askimagesize();
	int w = imw / decimate_;
	int h = imh / decimate_;
	converter_ = RGB_Converter_422::instance();
	setdibsize(w, h);
	set_size_422(w, h);
	allocref();
}

void Win411Grabber::setsize()
{
	askimagesize();
	int w = imw / decimate_;
	int h = imh / decimate_;
	converter_ = RGB_Converter_411::instance();
	setdibsize(w, h);
	set_size_411(w, h);
	allocref();
}

void WinCIFGrabber::setsize()
{
	int w = 2 * 352 / decimate_;
	int h = 2 * 288 / decimate_;
	converter_ = RGB_Converter_411::instance();
	setdibsize(w, h);
	set_size_411(w, h);
	allocref();
}

